import logging
import inspect
import threading
from datetime import datetime

import os
import traceback
import time


# Configure basic logging settings
def configure_logging(filename="agents_app.log", level=logging.DEBUG):
    logging.basicConfig(
        filename=f"log/{filename}_{datetime.now():%Y-%m-%d_%H-%M-%S}.log",
        level=level,
        format="%(asctime)s - %(levelname)s - %(message)s",
        filemode="w",
    )
    print(f'logfile:: {filename}')

def getTracebackWithReturn():
    #if int(os.environ['G_TRACEBACK_FLAG']) == 1 :
    #traceback.print_stack()
    retString=[]
    for line in traceback.format_stack():
        retString.append(line.strip()) 
    print(retString)     
    return retString



# Simplified logger function
def log(message, level=logging.INFO):
    """Logs a message with the specified level."""
    frame = inspect.currentframe().f_back
    thread_id = threading.get_ident()
    logger = logging.getLogger()
    msg = f'Thread ID: {thread_id} - {frame.f_globals["__name__"]}:{frame.f_lineno} - {message}'
    if level == logging.DEBUG:
        logger.debug(msg)
    elif level == logging.INFO:
        logger.info(msg)
    elif level == logging.WARNING:
        logger.warning(msg)
    elif level == logging.ERROR:
        logger.error(msg)
    elif level == logging.CRITICAL:
        logger.critical(msg)
    else:
        logger.log(level, msg)


# # Example usage
# if __name__ == "__main__":
#     configure_logging("my_app", logging.DEBUG)
#     log("This is a debug message.", logging.DEBUG)
#     log("This is an info message.")
#     log("This is a warning message.", logging.WARNING)
#     log("This is an error message.", logging.ERROR)
#     log("This is a critical message.", logging.CRITICAL)



from typing import Callable, Type
import inspect


def extract_func_params(func: Callable):
    funcs_signature = inspect.signature(func)
    funcs_params = funcs_signature.parameters
    return {
        name: param.default != inspect.Parameter.empty
        for name, param in funcs_params.items()
        if name != "self"  # noqa: E501
    }  # {"<attr_name>":"<default_is_empty>"}


def extract_class_init_attrs(clss: Type):
    return extract_func_params(clss.__init__)


def extract_cls_method_params(clss: Type, method: Callable):
    return extract_func_params(getattr(clss, method))



from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class GithubInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class GithubOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by Github.")


class GithubSearchTool(BaseTool):
    name: str = "GithubSearch Tool"
    description: str = (
        "A tool which can be used to retrieve information regarding respective repository like code changes, commits, active PRs, issues, etc by using natural language."
    )

    @tool(args_schema=GithubInputSchema, output_schema=GithubOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import github_toolkit

        return github_toolkit(searchString=search_str)



from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class GmailInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class GmailOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by gmail.")


class GmailSearchTool(BaseTool):
    name: str = "GmailSearch Tool"
    description: str = (
        "A tool which can be used to perform actions on gmail by using natural language"
    )

    @tool(args_schema=GmailInputSchema, output_schema=GmailOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import gmail_toolkit

        return gmail_toolkit(searchString=search_str)



from genai_agents.tools.base import BaseTool,tool
from pydantic import BaseModel, Field

class YoutubeSearchInputSchema(BaseModel):
    search_str: str = Field(description="Query used to search videos for on Youtube")

class YoutubeSearchOutputSchema(BaseModel):
    response: str = Field(description="Response from the YoutubeSearch tool.")

class YoutubeSearchTool(BaseTool):
    name: str = "YoutubeSearch Tool"
    description: str = (
        "A tool that can be used to search for videos on YouTube based on specific queries."
    )
    @tool(args_schema=YoutubeSearchInputSchema, output_schema=YoutubeSearchOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import getYoutubeSearchResults

        getYoutubeSearchResults(searchString=search_str)


import sys
import threading
import logging

from genai_agents.AIAgent import (
    createAndsendMessageProfTrigger,
    getGMapper,
    main_condition,
    setGMapper,
    setGTimerList,
    waitonConditionMain,
)
from genai_agents.MessageBroker import NameIndexMapper


def agentThreadsStart(agentObjectList):
    mapper = getGMapper()
    for obj in agentObjectList:
        obj.start_agent(mapper)


def agentInitSystem(agent_list):
    logging.debug("kick-off of the program")
    mapper = NameIndexMapper()
    timerPool = mapper.timerPool
    setGTimerList(timerPool)

    for agent in agent_list:
        mapper.add_mapping(agent)

    setGMapper(mapper)
    timer_thread = threading.Thread(target=timerPool.run)
    timer_thread.daemon = True
    logging.info("timer demon started")
    timer_thread.start()

def triggerAgent(agent_list, godTimerDuration):
    msg1 = "HGI sending trigger message to start the execution"
    mapper = getGMapper()
    for agent in agent_list:
        createAndsendMessageProfTrigger("profAgent", agent, msg1, mapper)
    waitonConditionMain(main_condition, godTimerDuration)
    logging.info("final exit")
    sys.exit()

def kickOffGenAIAgents2(AgentObjects, triggerAgentObjectsList): 
    agent_list=[]

    #extract agent names
    for agentObj in AgentObjects:
        agent_list.append(agentObj.agentName)
    agentInitSystem(agent_list=agent_list)
    print(agent_list)
    agentThreadsStart(agentObjectList=AgentObjects)

    #extract - trigger agent list
    triggerAgentList=[]
    for triggerObj in triggerAgentObjectsList:
        triggerAgentList.append(triggerObj.agentName)
    print(triggerAgentList)
    triggerAgent(triggerAgentList, godTimerDuration=10000)

def searchItemInList(DynamicAgentObjectsList, target):
    for  item in DynamicAgentObjectsList:
        if item.agentName == target.agentName:
            # print("Found")
            return True       
    return False
    
def kickOffGenAIAgents(AgentObjects, triggerAgentObjectsList, DynamicAgentObjectsList=[]): 
    agent_list=[]
    #extract agent names
    for agentObj in AgentObjects:
        agent_list.append(agentObj.agentName)
    agentInitSystem(agent_list=agent_list)
    print(agent_list)
    if(not DynamicAgentObjectsList):
        agentThreadsStart(agentObjectList=AgentObjects)
    else:
        for item in AgentObjects:
            if(not searchItemInList(DynamicAgentObjectsList, item)):
                # print("Not Found")
                mapper = getGMapper()
                item.start_agent(mapper)
    #extract - trigger agent list
    triggerAgentList=[]
    for triggerObj in triggerAgentObjectsList:
        triggerAgentList.append(triggerObj.agentName)
    print(triggerAgentList)
    triggerAgent(triggerAgentList, godTimerDuration=10000)


from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class SerperSpecificInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class SerperSpecificOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by SerperSpecific.")


class SerperSpecificSearchTool(BaseTool):
    name: str = "SerperSpecificSearch Tool"
    description: str = (
        "A tool which can be used to scrape information from the search engine for a specific type and period using natural language"
    )

    @tool(args_schema=SerperSpecificInputSchema, output_schema=SerperSpecificOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import getSerperScrapeForSpecificTypeAndPeriod

        return getSerperScrapeForSpecificTypeAndPeriod(searchString=search_str)



import logging
import threading
import time
from concurrent.futures import ThreadPoolExecutor

import spacy

from genai_agents.utils.llmTasks import handleLLMTask, tools_handler
from genai_agents.MessageBroker import NameIndexMapper
from genai_agents.PQ import Message, MessageType, MultiThreadPriorityQueue
from genai_agents.tools.tools_utils import getToolExecutionResults

from genai_agents.tools.tools_db import (
    getDuckduckgoSearchResults,
    TOOLS_DICT,
    TOOLS_DICT_MAPPING,
)
from genai_agents.tools.tools_utils import search_string_in_list
from spacytextblob.spacytextblob import SpacyTextBlob

g_mapper = None
g_timerlist = None
main_condition = threading.Condition()


def createDynamicAgent(agent, *functions):
    for startAgent, mapper in functions:
        startAgent(agent, mapper)


def setGMapper(mapper):
    global g_mapper
    g_mapper = mapper
    logging.debug(f"mapper object = {g_mapper.__dict__}")


def getGMapper():
    logging.debug(f"mapper object from get = {g_mapper.__dict__}")
    return g_mapper


def setGTimerList(timerPool):
    global g_timerlist
    g_timerlist = timerPool


def getGTimerList(timerPool):
    return g_timerlist


def waitonConditionAgent(mapper, agentName, duration):
    lindex = mapper.get_index_by_name(agentName)
    logging.debug(f"index of {agentName} is {lindex}")
    cond = mapper.COND_LIST[lindex]
    cond.acquire()
    cond.wait(timeout=duration)
    cond.release()


def wakeUpAgent(mapper, agentName):
    lindex = mapper.get_index_by_name(agentName)
    cond = mapper.COND_LIST[lindex]
    cond.acquire()
    cond.notify()
    cond.release()


def waitonConditionMain(condition, duration):
    condition.acquire()
    condition.wait(timeout=duration)
    condition.release()


def wakeupMainForKill(condition):
    condition.acquire()
    condition.notify()
    condition.release()


def createAndsendMessage(srcAgent, dstAgent, body, mapper):
    priority = 0
    perception_type = 0
    message_type = MessageType.AGENT_MSG
    body = body
    message = Message(priority, perception_type, message_type, srcAgent, dstAgent, body)
    pq = mapper.get_PQ_by_name(dstAgent)
    time.sleep(10)
    pq.insert(message)
    wakeUpAgent(mapper, dstAgent)


def createAndsendMessageProfTrigger(srcAgent, dstAgent, body, mapper):
    priority = 0
    perception_type = 0
    message_type = MessageType.PROF_AGENT_TRIGGER
    body = body
    message = Message(priority, perception_type, message_type, srcAgent, dstAgent, body)
    pq = mapper.get_PQ_by_name(dstAgent)
    time.sleep(5)
    pq.insert(message)
    wakeUpAgent(mapper, dstAgent)


def evalLLMResponse(resp):
    print(f"evalLLMResponse::{resp}")
    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe("spacytextblob")
    doc = nlp(resp)
    polarity = doc._.blob.polarity
    print(f"sentiment {polarity}")
    if polarity > 0:
        return True
    return False


def createAndSendFeedbackMessage(srcAgent, dstAgent, body, mapper, feedbackSummary):
    priority = 0
    perception_type = 0
    message_type = MessageType.FEEDBACK_MSG
    final_reponse = "revise" + body + "based on feedback " + body
    logging.debug(
        f"the body of feedback messgae...............................{final_reponse}",
    )
    message = Message(
        priority, perception_type, message_type, srcAgent, dstAgent, final_reponse
    )
    if not dstAgent == "profAgent":
        pq = mapper.get_PQ_by_name(dstAgent)
        time.sleep(10)
        pq.insert(message)
        wakeUpAgent(mapper, dstAgent)
        return
    else:
        logging.debug(f"response is going to profAgent f{message}")
        return


def example_callbackTimer(dstAgent, timerName, timerValue):
    logging.debug(f"Callback executed at {dstAgent}:{timerName}:{timerValue}")
    priority = 0
    perception_type = 0
    message_type = MessageType.TIMER_MSG
    body = "This is time out message"
    message = Message(
        priority, perception_type, message_type, "TIMER_AGENT", dstAgent, body
    )
    mapper = getGMapper()

    pq = mapper.get_PQ_by_name(dstAgent)
    pq.insert(message)
    logging.debug(f"timer message is inserted for {dstAgent}  {mapper.__dict__}")
    wakeUpAgent(mapper, dstAgent)


class AIAgent:
    def __init__(
        self,
        agentName,
        role,
        goal,
        backstory,
        capability,
        task,
        output_consumer_agent,
        llm_api="azure",
        tools_list=[],
        feedback=False,
        agent_type="STATIC",
        multiplicity=0,
        aggregator=0,
        onAggregationAction=None,
        creator=None,
        HGI_Intf=None,
        llm_resp_timer_value=20000,
    ):
        self.agentName = agentName
        self.aggregator = aggregator
        self.onAggregationAction = onAggregationAction
        self.creator = creator
        self.createAgentFlag = 0
        self.pq = MultiThreadPriorityQueue()
        self.condition = threading.Condition()
        self.is_data_found = False
        self.executor = ThreadPoolExecutor(
            max_workers=2
        )  # Adjust max_workers based on your requirement
        self.agent_thread = []
        self.mapper = []
        self.execute_task = "perform the task"
        self.consumerAgent = output_consumer_agent
        self.timerNameLLM = "LLM_TIMER"
        self.NoOfFeedbackMsgs = 0
        self.MAX_NoOfFeedbackMsgs = 1
        self.role = role
        self.feedback = feedback
        self.goal = goal
        self.backstory = backstory
        self.capability = capability
        self.agent_type = agent_type
        self.multiplicity = multiplicity
        self.task = task
        self.output_consumer_agent = output_consumer_agent
        self.HGI_Intf = HGI_Intf
        self.llm_api = llm_api
        self.timerValueLLM = llm_resp_timer_value
        self.tools_list = tools_list
        self.FeedBackMsgReceived = False

    def format_input(self, input_data):
        # Implement the logic to format input data
        pass

    def format_output(self, output_data):
        # Implement the logic to format output data
        pass

    def execute(self, messageList):
        # Implement the logic for the execution of the task
        message = messageList[0]
        if self.aggregator > 0:
            # overwrite the body of the first message with aggregated content
            message.body = self.onAggregationAction(
                self.agentName, self.output_consumer_agent, None, messageList
            )

        consumer = self.consumerAgent
        resp = f"sending mesage from {self.agentName}"

        if self.tools_list:
            print("Using tools")
            _tools = []

            for user_tool in self.tools_list:
                tool_name = user_tool.name  # Assuming user_tool has a name attribute
                if tool_name in TOOLS_DICT_MAPPING:
                    # If the tool exists in TOOLS_DICT, append its info
                    _tools.append(TOOLS_DICT_MAPPING[tool_name])
                else:
                    # If the tool is not found, append a custom tool dictionary
                    _tools.append(
                        {
                            "category": "custom",
                            **user_tool.get_tool_info(),
                        }
                    )

            if self.tools_list:
                resp = tools_handler(
                    tools=_tools,
                    task_input={
                        "role": self.role,
                        "name": self.agentName,
                        "backstory": self.backstory,
                        "task": self.task,
                        "goal": self.goal,
                        "capability": self.capability,
                        "instructions": self.task,  # TODO: To be Removed or fixed as its redundant
                    },
                )
                logging.info(f"resp from tools_handler : {resp}")

        elif self.capability == "search_executor":
            indices = search_string_in_list(self.tools_list, "duckduckgo-search")
            if indices:
                results = getDuckduckgoSearchResults(self.goal + self.task)
                print(f"executed duckduck search and results :{results}")
                resp = f" {results}"

        elif (
            self.capability == "llm_task_executor"
        ):  # and (self.role=="SUMMARISER" or self.role=="EMAILER" or self.role=="Coder" or self.role=="Reviewer")):
            print(f"\n      role being performed is { self.role } \n       ")

            if not self.feedback or self.NoOfFeedbackMsgs < self.MAX_NoOfFeedbackMsgs:
                resp = handleLLMTask(
                    message.body,
                    self.role,
                    self.backstory,
                    self.goal,
                    self.task,
                    self.llm_api,
                )
            else:
                logging.info("response for  final feedback ")
                resp = message.body

        logging.debug(f"response of agent {self.agentName} is { resp}")
        # STOP THE TIMER
        self.mapper.timerStopMapper(
            agentName=self.agentName, timerName=self.timerNameLLM
        )
        # feedback is supported for only non aggregator agents
        if (
            self.feedback
            and self.aggregator == 0
            and self.NoOfFeedbackMsgs < self.MAX_NoOfFeedbackMsgs
        ):
            status = evalLLMResponse(resp)
            if not status:
                self.NoOfFeedbackMsgs += 1
                createAndSendFeedbackMessage(
                    self.agentName, message.srcAgent, message.body, self.mapper, resp
                )
                return
            else:
                logging.debug("feedback on message is positive")
                resp = message.body  # update the body with  original of the agent

        if self.HGI_Intf:
            logging.debug('...............executing human tool.....................\n')
            resp, feedback, actionself = self.HGI_Intf(self.agentName, resp, self.consumerAgent)
            logging.debug(f' the response from human tool:: {resp}')

        if not consumer or consumer == "HGI":
            logging.debug(f"final message to {consumer}= {resp}")
            print(f"final OUTPUT for {consumer} ::{resp}")
            logging.debug("before wake for exit..........")
            wakeupMainForKill(main_condition)

        if consumer and consumer != "HGI":
            logging.debug(f"{self.agentName}..to.. {consumer} content:: {resp}")
            createAndsendMessage(self.agentName, consumer, resp, self.mapper)

    def run(self):
        myname = self.agentName
        logging.debug(f"I am agent {myname}")

        multiplicity = self.multiplicity
        msgList = []
        while True:
            message = self.pq.retrieve_first_message()

            if message:
                msgList.append(message)
                logging.debug(f"agent {myname} retrieved message: {message.__dict__}")
                # print(message.__dict__)
                logging.debug("appended received message to list for aggregation")
                if len(msgList) == self.aggregator or self.aggregator == 0:
                    if self.creator and self.createAgentFlag == 0:
                        logging.debug(
                            f"creating dynamic agent {self.output_consumer_agent} in {self.agentName}",
                        )
                        functions_to_execute = [(AIAgent.start_agent, self.mapper)]
                        createDynamicAgent(self.creator, *functions_to_execute)
                        self.createAgentFlag = 1
                    self.mapper.timerStartMapper(
                        agentName=self.agentName,
                        timerName=self.timerNameLLM,
                        callback=example_callbackTimer,
                        timervalue=self.timerValueLLM,
                    )
                    self.execute(msgList)
                    msgList = []
            else:
                logging.debug(
                    f"{myname} agent found no messages from PQ .. going to wait."
                )
                waitonConditionAgent(self.mapper, self.agentName, 100)

    def start_agent(self, mapper):
        # Start the agent in a new thread
        self.mapper = mapper  # python - call by value or reference????
        self.pq = mapper.get_PQ_by_name(self.agentName)
        self.condition = mapper.get_COND_by_name(self.agentName)
        self.agent_thread = threading.Thread(target=self.run)
        self.agent_thread.daemon = True
        self.agent_thread.start()


# Example Usage:
if __name__ == "__main__":
    # Example usage:

    # Start the agent in a new thread
    agent_list = ["RESEARCHER", "WRITER", "EMAILER"]
    mapper = NameIndexMapper()
    mapper.add_mapping(agent_list[0])
    mapper.add_mapping(agent_list[1])
    mapper.add_mapping(agent_list[2])
    agent1 = AIAgent(agentName=agent_list[0], consumerAgent=agent_list[1])
    agent2 = AIAgent(agentName=agent_list[1], consumerAgent=agent_list[2])
    agent3 = AIAgent(agentName=agent_list[2], consumerAgent=agent_list[0])

    # , role="RESEARCHER", goal="search for latest trends in Carona treatment", capability="INTERNET_SEARCHER", agent_type="STATIC",
    #                multiplicity=3, task="search internet for the goal", output_consumer_agent=agent_list[1],
    #               llm_api="gpt4", go_d_timer=10000, llm_resp_timer=20, tools_list=["Google", "Bing","DuckduckGo-search" ])

    agent1.start_agent(mapper)
    agent2.start_agent(mapper)
    agent3.start_agent(mapper)
    time.sleep(5)  # Simulating data availability after 5 seconds
    msg1 = "message for agent RESEARCHER from ENVIRONMENT "
    createAndsendMessage("ProfAgent", agent_list[0], msg1, mapper)



import logging
import queue
import random
import threading
import time
from enum import Enum


class MessageType(Enum):
    TIMER_MSG = 1
    PROF_AGENT_TRIGGER = 2
    STOP_MSG = 3
    SHUTDOWN_MSG = 4
    AGENT_MSG = 5
    FEEDBACK_MSG = 6
    STATUS_RSP = 7
    LLM_CHANGE_MSG = 8


class Message:
    def __init__(
        self, priority, perception_type, message_type, srcAgent, dstAgent, body
    ):
        self.priority = priority
        self.perception_type = perception_type
        self.message_type = message_type
        self.srcAgent = srcAgent
        self.dstAgent = dstAgent
        self.body = body
        logging.debug(
            f"message contents:{priority}:{perception_type}::{message_type} ::{srcAgent} ->{dstAgent} :{body}"
        )

    def __lt__(self, other):
        return self.priority < other.priority


class MultiThreadPriorityQueue:
    def __init__(self):
        self.priority_queue = queue.PriorityQueue()
        self.lock = threading.Lock()

    def insert(self, message):
        with self.lock:
            self.priority_queue.put(message)

    def retrieve_first_message(self):
        with self.lock:
            if not self.priority_queue.empty():
                return self.priority_queue.get()
            else:
                return None

    def list_all_elements(self):
        with self.lock:
            return list(self.priority_queue.queue)


def producer(queue):
    for _ in range(5):
        priority = random.randint(1, 10)
        perception_type = 0
        message_type = random.choice(list(MessageType))
        body = f"Body_{random.randint(100, 999)}"
        message = Message(priority, perception_type, message_type, "", "", body)
        queue.insert(message)
        time.sleep(random.uniform(0.1, 0.5))


def consumer(queue):
    for _ in range(5):
        message = queue.retrieve_first_message()
        if message:
            logging.debug(f"Consumer retrieved message: {message.__dict__}")
        else:
            logging.info("Consumer found no messages.")
        time.sleep(10)


if __name__ == "__main__":
    mt_priority_queue = MultiThreadPriorityQueue()

    # Create producer and consumer threads
    producer_thread = threading.Thread(target=producer, args=(mt_priority_queue,))
    consumer_thread = threading.Thread(target=consumer, args=(mt_priority_queue,))

    # Start the threads
    producer_thread.start()
    consumer_thread.start()

    # Wait for threads to finish
    producer_thread.join()
    consumer_thread.join()

    # List all elements in the queue after threads finish
    all_elements = mt_priority_queue.list_all_elements()
    logging.info("\nAll Elements in Priority Queue:")
    for element in all_elements:
        logging.info(element.__dict__)



from genai_agents.tools.base import BaseTool,tool
from pydantic import BaseModel, Field

class WikipediaToolInputSchema(BaseModel):
    search_str: str = Field(description="Query used to search the Wikipedia API")

class WikipediaToolOutputSchema(BaseModel):
    response: str = Field(description="Response from the Wikipedia tool.")

class WikipediaTool(BaseTool):
    name: str = "Wikipedia Tool"
    description: str = (
        "A tool designed to Tool that searches the Wikipedia API for a specific query."
    )
    @tool(args_schema=WikipediaToolInputSchema, output_schema=WikipediaToolOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import wikipediaTool

        wikipediaTool(searchString=search_str)



import os
from langchain_openai import AzureChatOpenAI
from langchain.agents import (
     AgentExecutor,
     AgentType,
     OpenAIFunctionsAgent,
     Tool,
     create_openai_functions_agent,
     initialize_agent,
     load_tools,
     tool,
)
from genai_agents.tools.base import BaseTool,tool
from pydantic import BaseModel, Field
from PIL import Image


#issues - not working
def DallEImageGenerator(inputString):
    llm_2 = AzureChatOpenAI(
    model_name = "gpt4-32k",
    openai_api_version="2023-05-15",
    azure_deployment="gpt4-inference"
    )
    llm=llm_2
    tools = load_tools(["dalle-image-generator"])
    agent = initialize_agent(tools, llm, agent="zero-shot-react-description", verbose=False)
    image = agent.run(inputString)
    return image





from datetime import datetime
import logging
from pathlib import Path

BASE_PATH = "logs"
pth = Path(BASE_PATH)
pth.mkdir(parents=True, exist_ok=True)

# Setup basic logging configuration
logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s %(levelname)s::%(thread)d::%(pathname)s:%(lineno)d:%(funcName)s:%(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    filename=f'{pth.absolute()}/application_{datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}.log',
    filemode="w",
)
logging.info(f"Logs stored at {pth.absolute()}")



from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class XorbitsInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class XorbitsOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by Xorbits.")


class XorbitsSearchTool(BaseTool):
    name: str = "XorbitsSearch Tool"
    description: str = (
        "A tool which can be used to retrieve information from files by performing pandas or numpy code by using natural language."
    )

    @tool(args_schema=XorbitsInputSchema, output_schema=XorbitsOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import xorbits_toolkit

        return xorbits_toolkit(searchString=search_str)



from typing import Callable, Any, Type
from pydantic import BaseModel, Field


def tool(
    args_schema: Type[BaseModel], output_schema: Type[BaseModel] = None
) -> Callable:
    def decorator(func: Callable) -> Callable:
        # Ensure we're accessing the correct attribute for Pydantic fields
        func.description = func.__doc__
        func.args = {
            field_name: field.description
            for field_name, field in args_schema.model_fields.items()
        }
        func.output_schema = {
            field_name: field.description
            for field_name, field in output_schema.model_fields.items()
        }

        def wrapper(self, *args, **kwargs) -> Any:
            return func(self, *args, **kwargs)

        # Attach metadata to the wrapper function to ensure it's accessible
        wrapper.args = func.args
        wrapper.output_schema = func.output_schema
        return wrapper

    return decorator


class BaseToolInputSchema(BaseModel):
    data: str = Field(description="Data to be passed")


class BaseToolOutputSchema(BaseModel):
    response: str = Field(description="Data returned from the tool")


class BaseTool:
    """Baseclass for tools"""

    name: str = "Hello"
    description: str = "Base tool description"

    @tool(args_schema=BaseToolInputSchema, output_schema=BaseToolOutputSchema)
    def _run(self, data: str):
        """Method description."""
        raise NotImplementedError

    @classmethod
    def get_tool_info(cls):
        tool_info = {
            "tool_name": cls.name,
            "description": cls.description,
            "args": cls._run.args if hasattr(cls._run, "args") else {},
            "output": (
                cls._run.output_schema if hasattr(cls._run, "output_schema") else None
            ),
            "cls": {
                "kls": cls.__name__,
                "module": cls.__module__,
            },
        }
        return tool_info



from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class DocumentCompareInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class DocumentCompareOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by DocumentCompare.")


class DocumentCompareSearchTool(BaseTool):
    name: str = "DocumentCompareSearch Tool"
    description: str = (
        "A tool which can be used to by the agent to question uploaded files by the user."
    )

    @tool(args_schema=DocumentCompareInputSchema, output_schema=DocumentCompareOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import DocuCompare
        
        return DocuCompare(searchString=search_str)


import logging
import threading
from genai_agents.PQ import MultiThreadPriorityQueue, producer, consumer
from genai_agents.timerPool import TimerList
from genai_agents.utils.yamlParse import getYamlAttribute
num_queues = getYamlAttribute('MAX_NUMBER_OF_AGENTS')


class NameIndexMapper:
    def __init__(self):
        self.name_to_index = {}
        self.index_to_name = {}
        self.PQ_LIST = [MultiThreadPriorityQueue() for i in range(num_queues)]
        self.COND_LIST = [threading.Condition() for i in range(num_queues)]
        self.lock = threading.Lock()
        self.timerPool = TimerList()
        logging.debug(f"created message broker infra with agents no {num_queues}")

    def add_mapping(self, name):
        with self.lock:
            if name not in self.name_to_index:
                index = len(self.name_to_index)
                self.name_to_index[name] = index
                self.index_to_name[index] = name

    def get_index_by_name(self, name):
        with self.lock:
            return self.name_to_index.get(name)

    def get_name_by_index(self, index):
        with self.lock:
            return self.index_to_name.get(index)

    def get_PQ_by_name(self, agentName):
        index = self.get_index_by_name(agentName)
        logging.debug(f"index of {agentName} is {index}")
        return self.PQ_LIST[index]

    def get_COND_by_name(self, agentName):
        index = self.get_index_by_name(agentName)
        logging.debug(f"index of {agentName} is {index}")
        return self.COND_LIST[index]

    def timerStartMapper(self, agentName, timerName, callback, timervalue):
        newTimer = self.timerPool.add_timer(agentName, timerName, callback, timervalue)
        return newTimer

    def timerStopMapper(self, agentName, timerName):
        self.timerPool.remove_timerWithNames(agentName, timerName)


# Example Usage:
if __name__ == "__main__":
    mapper = NameIndexMapper()

    # Adding mappings
    mapper.add_mapping("Alice")
    mapper.add_mapping("Bob")
    mapper.add_mapping("Charlie")

    # Using the mapper
    index_for_alice = mapper.get_index_by_name("Alice")
    name_for_index_1 = mapper.get_name_by_index(1)

    logging.info(f"Index for Alice: {index_for_alice}")
    logging.info(f"Name for index 1: {name_for_index_1}")

    PQ1 = mapper.get_PQ_by_name("Alice")
    PQ2 = mapper.get_PQ_by_name("Bob")
    PQ3 = mapper.get_PQ_by_name("Charlie")
    # mapper.sleepAgent("Alice",100)
    # mapper.sleepAgent("Bob",100)
    # mapper.sleepAgent("Charlie",100)
    # mapper.wakeupAgent("Alice")
    # mapper.wakeupAgent("Bob")
    # mapper.wakeupAgent("Charlie")

    producer_thread = threading.Thread(target=producer, args=(PQ1,))
    producer_thread = threading.Thread(target=producer, args=(PQ2,))
    producer_thread = threading.Thread(target=producer, args=(PQ3,))
    consumer_thread = threading.Thread(target=consumer, args=(PQ1,))
    consumer_thread = threading.Thread(target=consumer, args=(PQ2,))
    consumer_thread = threading.Thread(target=consumer, args=(PQ3,))

    # Start the threads
    producer_thread.start()
    consumer_thread.start()

    # Wait for threads to finish
    producer_thread.join()
    consumer_thread.join()

    # List all elements in the queue after threads finish
    all_elements = PQ1.list_all_elements()
    logging.info("\nAll Elements in Priority Queue1:")
    for element in all_elements:
        logging.info(element.__dict__)

    all_elements = PQ3.list_all_elements()
    logging.info("\nAll Elements in Priority Queue3:")
    for element in all_elements:
        logging.info(element.__dict__)

    all_elements = PQ2.list_all_elements()
    logging.info("\nAll Elements in Priority Queue2:")
    for element in all_elements:
        logging.info(element.__dict__)



import logging
from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field
from genai_agents.custom_tools.custom_tool_db import proxyToolkit

class ProxyInputSchema(BaseModel):
    input_str: str = Field(description="Input to be returned.")


class ProxyOutputSchema(BaseModel):
    response: str = Field(description="Returns the input.")


class ProxyTool(BaseTool):
    name: str = "Proxy Tool"
    description: str = "A tool that returns the input."

    @tool(args_schema=ProxyInputSchema, output_schema=ProxyOutputSchema)
    def _run(self, input_str: str = None):
        logging.debug(f"Proxy tool called with some input of length {len(input_str)}.")
        print("custom tool executed")
        return proxyToolkit(input_str)



from genai_agents.tools.base import BaseTool,tool
from pydantic import BaseModel, Field

class GoogleSerperSearchInputSchema(BaseModel):
    search_str: str = Field(description="Query used to search online from Google")

class GoogleSerperSearchOutputSchema(BaseModel):
    response: str = Field(description="Response from the YoutubeSearch tool.")

class GoogleSerperSearchTool(BaseTool):
    name: str = "GoogleSerperSearch Tool"
    description: str = (
        "Tool used to perform a search online using a Google SERP (Search Engine Results Page) scraping API wrapper. "
    )
    @tool(args_schema=GoogleSerperSearchInputSchema, output_schema=GoogleSerperSearchOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import getSerperScrape
        getSerperScrape(searchString=search_str)


from genai_agents.tools.functional_tools.github_tool import GithubSearchTool
from genai_agents.tools.functional_tools.gmail_tool import GmailSearchTool
from genai_agents.tools.functional_tools.documentcompare_tool import DocumentCompareSearchTool
from genai_agents.tools.functional_tools.nasa_tool import NasaSearchTool
from genai_agents.tools.functional_tools.openweathermap_tool import OpenWeatherMapSearchTool
from genai_agents.tools.functional_tools.polygon_tool import PolygonSearchTool
from genai_agents.tools.functional_tools.sql_tool import SqlSearchTool
from genai_agents.tools.functional_tools.xorbits_tool import XorbitsSearchTool
from genai_agents.tools.functional_tools.exaSearch_tool import exaSearchTool


from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class PolygonInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class PolygonOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by Polygon.")


class PolygonSearchTool(BaseTool):
    name: str = "PolygonSearch Tool"
    description: str = (
        "A tool which can be used to retrieve information regarding current Stock prices and their historical data using natural language."
    )

    @tool(args_schema=PolygonInputSchema, output_schema=PolygonOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import polygon_toolkit
        return polygon_toolkit(searchString=search_str)


from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class OpenWeatherMapInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class OpenWeatherMapOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by OpenWeatherMap.")


class OpenWeatherMapSearchTool(BaseTool):
    name: str = "OpenWeatherMapSearch Tool"
    description: str = (
        "A tool which can be used to retrieve information about weather of any place in the world."
    )

    @tool(args_schema=OpenWeatherMapInputSchema, output_schema=OpenWeatherMapOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import open_weather_app

        return open_weather_app(searchString=search_str)




from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class GoogleFinanceStockInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class GoogleFinanceStockOutputSchema(BaseModel):
    response: str = Field(description="Response from the GoogleFinanceStockSearchtool engine.")


class GoogleFinanceSearchTool(BaseTool):
    name: str = "GoogleFinanceStockSearch Tool"
    description: str = (
        "A tool that uses the Google Finance Tool to get information from the Google Finance page"
    )
    @tool(args_schema=GoogleFinanceStockInputSchema, output_schema=GoogleFinanceStockOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import googleFinanceStockSearch
        googleFinanceStockSearch(searchString=search_str)  


# #not working
# from langchain.agents import (
#      AgentExecutor,
#      AgentType,
#      OpenAIFunctionsAgent,
#      Tool,
#      create_openai_functions_agent,
#      initialize_agent,
#      load_tools,
#      tool,
# )
# import os
# import pprint

# from langchain_openai import AzureChatOpenAI

# os.environ["SERPAPI_API_KEY"]="09e27394ffb11f7a9d01e7eaf4204191a68c5dc633e4dd17a76b4c739472aead"
# os.environ["SERP_API_KEY"]="09e27394ffb11f7a9d01e7eaf4204191a68c5dc633e4dd17a76b4c739472aead"
# os.environ["AZURE_OPENAI_API_KEY"] = "a20bc67dbd7c47ed8c978bbcfdacf930"
# os.environ["AZURE_OPENAI_ENDPOINT"] = "https://gpt-res.openai.azure.com/"

# llm_1 = AzureChatOpenAI(
# model_name = "gpt4-32k",
# openai_api_version="2023-05-15",
# azure_deployment="gpt4-inference"
# )
# def googleFinanceStockSearch(inputString):
#     llm = llm_1 #OpenAI(temperature=0)
    
#     tools = load_tools(["google-scholar", "google-finance"], llm=llm)
#     agent = initialize_agent(
#     tools, llm, agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION, verbose=False
#     )
#     results= agent.run(inputString)
#     finalResult= pprint.pp(results)
#     print(finalResult)
#     return results

# print(googleFinanceStockSearch("What is google's current stock"))



import os
from pathlib import Path

import yaml

GENAIAGENTS_CONFIG_PATH = os.environ.get("GENAIAGENTS_CONFIG_PATH")
if not GENAIAGENTS_CONFIG_PATH:
    raise FileNotFoundError(
        "Environment variable not set: `GENAIAGENTS_CONFIG_PATH`"
    )
GENAIAGENTS_CONFIG_PATH = Path(GENAIAGENTS_CONFIG_PATH)
if not GENAIAGENTS_CONFIG_PATH.is_file():
    raise FileNotFoundError(
        f"No such file or directory: `{GENAIAGENTS_CONFIG_PATH.absolute()}`"
    )


def getYamlAttribute(attrName):
    with open(GENAIAGENTS_CONFIG_PATH, "r") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
        attrValue = data.get(attrName)
        return attrValue



import threading
import logging
import time

condition = threading.Condition()


def waitonCondition(condition, duration):
    condition.acquire()
    condition.wait(timeout=duration)
    condition.release()


def execute_function():
    logging.debug("Executing the function...")


def wakeupTimerThread():
    condition.acquire()
    condition.notify()
    condition.release()


class TimerNode:
    def __init__(self, agentName, timerName, callback, expiry_time):
        self.callback = callback
        self.expiry_time = expiry_time
        self.agentName = agentName
        self.timerName = timerName
        self.next = None
        self.prev = None


class TimerList:
    def __init__(self):
        self.head = None
        self.lock = threading.Lock()

    def add_timer(self, agentName, timerName, callback, interval):
        expiry_time = time.time() + interval
        logging.debug(f"current time: {time.time()} : internal {interval}")
        new_timer = TimerNode(agentName, timerName, callback, expiry_time)

        with self.lock:
            if not self.head or self.head.expiry_time > expiry_time:
                logging.debug("first timer inserted....")
                new_timer.next = self.head
                if self.head:
                    self.head.prev = new_timer
                self.head = new_timer
            # wakeupTimerThread()

            else:
                current = self.head
                while current.next and current.next.expiry_time <= expiry_time:
                    current = current.next

                new_timer.next = current.next
                if current.next:
                    current.next.prev = new_timer
                current.next = new_timer
                new_timer.prev = current
        wakeupTimerThread()
        return new_timer

    def remove_timer(self, timer_node):
        with self.lock:
            if timer_node.prev:
                timer_node.prev.next = timer_node.next
            else:
                self.head = timer_node.next

            if timer_node.next:
                timer_node.next.prev = timer_node.prev
            wakeupTimerThread()

    def remove_timerWithLock(self, timer_node):
        if timer_node.prev:
            timer_node.prev.next = timer_node.next
        else:
            self.head = timer_node.next

        if timer_node.next:
            timer_node.next.prev = timer_node.prev
        wakeupTimerThread()

    def remove_timerWithNames(self, agentName, timerName):
        logging.debug(f"remove timer is called for {agentName} {timerName}")
        with self.lock:
            if self.head:
                current = self.head
                while current:
                    if (
                        current.agentName == agentName
                        and current.timerName == timerName
                    ):
                        logging.debug(
                            f"timer found for agent {agentName} timer name {timerName}  for removal"
                        )
                        self.remove_timerWithLock(current)
                        return True
                    else:
                        current = current.next

            logging.debug(
                f"timer of agent {agentName} timer name {timerName} not found "
            )
            return False

    def list_active_timers(self):
        with self.lock:
            current = self.head
            while current:
                print(
                    f"{current.agentName}::Timer :{current.timerName}expires at: {current.expiry_time} "
                )
                current = current.next

    def run(self):
        logging.debug("timer thread started..................")
        while True:
            with self.lock:
                if self.head:
                    current_time = time.time()
                    time_to_wait = max(0, self.head.expiry_time - current_time)
                else:
                    time_to_wait = 1000  # float('inf')
            logging.debug(f"time to sleep: {time_to_wait}")
            waitonCondition(condition, time_to_wait)
            # time.sleep(time_to_wait)
            # print("sleep is over....")
            with self.lock:
                if self.head and time.time() >= self.head.expiry_time:
                    print("timer call to be called as timer expired...")
                    callback = self.head.callback
                    agentName = self.head.agentName
                    timerName = self.head.timerName
                    timerValue = self.head.expiry_time
                    # print("timer to be revomed....")
                    self.remove_timerWithLock(self.head)
                    # print("timer is removed......")

                    if callback:
                        logging.debug(
                            f"Callback executed at {agentName}:{timerName}:{timerValue}"
                        )
                        callback(agentName, timerName, timerValue)


def example_callback(agentName, timerName, timerValue):
    logging.debug(f"Callback executed at {agentName}:{timerName}:{timerValue}")


def timerStart(agentName, timerName, callback, timervalue):
    newTimer = timer_list.add_timer(agentName, timerName, callback, timervalue)


if __name__ == "__main__":
    timer_list = TimerList()

    # Create two threads to create multiple timers

    # thread1 = threading.Thread(target=lambda: timer_list.add_timer("thread1", example_callback, 13))
    # thread2 = threading.Thread(target=lambda: timer_list.add_timer("thread2", example_callback, 15))

    # Create the worker thread and pass the condition and duration to it
    # timer_thread = threading.Thread(target=timer_list.run, args=(condition, wait_duration))
    # Start the worker thread
    # Start the timer thread
    timer_thread = threading.Thread(target=timer_list.run)
    timer_thread.start()

    time.sleep(1)
    for i in range(50):
        timerStart("agent1", "tmr" + str(i), example_callback, 10 + i)
        timerStart("agent2", "tmr" + str(i), example_callback, 11 + i)
    print("Active Timers:")
    timer_list.list_active_timers()
    # condition.notify()
    # Start the timer-setting threads
    # thread1.start()
    # thread2.start()
    # Keep the main thread alive
    try:
        time.sleep(100)

    except KeyboardInterrupt:
        print("exception.....")
        # If the user interrupts the program, stop all threads
        timer_thread.join()
        # thread1.join()
        # thread2.join()



from genai_agents.tools.websearch.duckducksearch import DuckDuckGoSearchTool
from genai_agents.tools.websearch.google import GoogleSerperSearchTool
from genai_agents.tools.websearch.youtubesearch import YoutubeSearchTool
from genai_agents.tools.websearch.google_serper_specific import SerperSpecificSearchTool
from genai_agents.tools.websearch.serper_intermediate import SerperIntermediateSearchTool
from genai_agents.tools.websearch.yahoofinancenews import YahooFinanceTool
from genai_agents.tools.websearch.wikipedia_tool import WikipediaTool
from genai_agents.tools.websearch.googlejobsearch import GoogleJobSearchTool
from genai_agents.tools.websearch.google_finance_search import GoogleFinanceSearchTool



from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class SerperIntermediateInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class SerperIntermediateOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by SerperIntermediate.")


class SerperIntermediateSearchTool(BaseTool):
    name: str = "SerperIntermediateSearch Tool"
    description: str = (
        "A tool which can be useful for when you need to ask with search."
    )

    @tool(args_schema=SerperIntermediateInputSchema, output_schema=SerperIntermediateOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import googleSerperSearchIntermediateQuestions

        return googleSerperSearchIntermediateQuestions(searchString=search_str)


from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class exaSearchInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class exaSearchOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by exaSearch.")


class exaSearchTool(BaseTool):
    name: str = "exaSearchSearch Tool"
    description: str = (
        "A tool which can be used to do a Exa Search."
    )

    @tool(args_schema=exaSearchInputSchema, output_schema=exaSearchOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import exaSearch

        return exaSearch(searchString=search_str)



from genai_agents.tools.base import BaseTool



class GoogleSerperSearchTool(BaseTool):
    name = "Google Serper Search"
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import getSerperScrape
        getSerperScrape(searchString=search_str)



from genai_agents.tools.base import BaseTool,tool
from pydantic import BaseModel, Field

class YahooFinanceInputSchema(BaseModel):
    search_str: str = Field(description="Query used to search for news on Yahoo Finance")

class YahooFinanceOutputSchema(BaseModel):
    response: str = Field(description="Response from the YahooFinance tool.")

class YahooFinanceTool(BaseTool):
    name: str = "Yahoo Finance News Tool"
    description: str = (
        "A tool designed to explore financial news articles on Yahoo Finance."
    )
    @tool(args_schema=YahooFinanceInputSchema, output_schema=YahooFinanceOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import yahooFinanceNews

        yahooFinanceNews(searchString=search_str)


from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class DuckDuctGoInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class DuckDuctGoOutputSchema(BaseModel):
    response: str = Field(description="Response from the DuckDuckGo search engine.")


class DuckDuckGoSearchTool(BaseTool):
    name: str = "DuckDuckGoSearch Tool"
    description: str = (
        "A tool that can be used to search for words, documents, images, videos, news, maps and text translation using the DuckDuckGo.com search engine. Downloading files and images to a local hard drive."
    )

    @tool(args_schema=DuckDuctGoInputSchema, output_schema=DuckDuctGoOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import getDuckduckgoSearchResults

        return getDuckduckgoSearchResults(searchString=search_str)


from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class GoogleJobsInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class GoogleJobsOutputSchema(BaseModel):
    response: str = Field(description="Response from the GoogleJobsQuery engine.")


class GoogleJobSearchTool(BaseTool):
    name: str = "GoogleJobsSearch Tool"
    description: str = (
        "A tool that can be used to access the GoogleJobs API and search for jobs online "
    )
    @tool(args_schema=GoogleJobsInputSchema, output_schema=GoogleJobsOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import googleJobsSearch
        googleJobsSearch(searchString=search_str)



import asyncio
import os
import pprint
import dotenv
import xorbits.pandas as pd
from exa_py import Exa
from genai_agents.utils.yamlParse import getYamlAttribute
from langchain import hub
from langchain.agents import (
     AgentExecutor,
     AgentType,
     OpenAIFunctionsAgent,
     Tool,
     create_openai_functions_agent,
     initialize_agent,
     load_tools,
     tool,
)
import logging
#from langchain.llms import CTransformers
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.chains import LLMChain, RetrievalQA
from langchain_text_splitters import CharacterTextSplitter
from langchain_community.agent_toolkits import (
     AzureCognitiveServicesToolkit,
     GmailToolkit,
     O365Toolkit,
     PlayWrightBrowserToolkit,
     PowerBIToolkit,
     create_pbi_agent,
     create_sql_agent,
)
from langchain_community.agent_toolkits.github.toolkit import GitHubToolkit
from langchain_community.agent_toolkits.nasa.toolkit import NasaToolkit
from langchain_community.agent_toolkits.polygon.toolkit import PolygonToolkit
from langchain_community.document_loaders import PyPDFLoader
from langchain_community.vectorstores import FAISS
from langchain_core.prompts import PromptTemplate
from langchain_openai import AzureChatOpenAI, AzureOpenAIEmbeddings, OpenAI, OpenAIEmbeddings
from langchain_community.llms import CTransformers
from langchain_community.tools import (
     DuckDuckGoSearchRun,
     ElevenLabsText2SpeechTool,
     WikipediaQueryRun,
     YouTubeSearchTool,
)
from langchain_community.tools.google_jobs import GoogleJobsQueryRun
from langchain_community.tools.playwright.utils import create_async_playwright_browser
from langchain_community.tools.yahoo_finance_news import YahooFinanceNewsTool
from langchain_community.utilities import GoogleSerperAPIWrapper, WikipediaAPIWrapper
from langchain_community.utilities.github import GitHubAPIWrapper
from langchain_community.utilities.google_jobs import GoogleJobsAPIWrapper
from langchain_community.utilities.nasa import NasaAPIWrapper
from langchain_community.utilities.polygon import PolygonAPIWrapper
from langchain_community.utilities.powerbi import PowerBIDataset
from langchain_community.utilities.sql_database import SQLDatabase
from genai_agents.tools.functional_tools import *
from langchain_core.messages import SystemMessage
from langchain_core.runnables import RunnableParallel, RunnablePassthrough
from langchain_exa import ExaSearchRetriever, TextContentsOptions
from genai_agents.tools.functional_tools import *
from langchain_experimental.agents.agent_toolkits import create_xorbits_agent
from genai_agents.tools.websearch import *
from genai_agents.utils.extraction import extract_cls_method_params
import os


dotenv.load_dotenv()

TOOLS_DICT = [
    {
        "category": "Search",
        **DuckDuckGoSearchTool.get_tool_info(),
        "output": "Search results after using the tool",
    },
    {
        "category": "gmail",
        **GmailSearchTool.get_tool_info(),
        "output": "Gmail results after using the tool.", 
    },
    {
        "category" : "Search",
        **GithubSearchTool.get_tool_info(),
        "output": "Github results after using the tool.",
    },
    {
        "category" : "Search",
        **YoutubeSearchTool.get_tool_info(),
        "output": "Youtube results after using the tool.",
    },
    {
        "category" : "Compare",
        **DocumentCompareSearchTool.get_tool_info(),
        "output": "Document comparison results after using the tool.",
    },
    {
        "category" : "Search",
        **NasaSearchTool.get_tool_info(),
        "output": "Nasa results after using the tool.",
    },
    {
        "category" : "Search",
        **OpenWeatherMapSearchTool.get_tool_info(),
        "output": "OpenWeatherMap results after using the tool.",
    },
    {
        "category" : "Search",
        **PolygonSearchTool.get_tool_info(),
        "output": "Polygon results after using the tool.",
    },
    {
        "category" : "Search",
        **SqlSearchTool.get_tool_info(),
        "output": "Sql results after using the tool.",
    },
    {
        "category" : "Search",
        **XorbitsSearchTool.get_tool_info(),
        "output": "Xorbits results after using the tool.",
    },
    {
        "category" : "Search",
        **GoogleSerperSearchTool.get_tool_info(),
        "output": "GoogleSerper results after using the tool.",
    },
    {
        "category" : "Search",
        **YoutubeSearchTool.get_tool_info(),
        "output": "Youtube results after using the tool.",
    },
    {
        "category" : "Search",
        **GoogleFinanceSearchTool.get_tool_info(),
        "output": "GoogleFinance results after using the tool.",
    },
    {
        "category" : "Search",
        **WikipediaTool.get_tool_info(),
        "output": "Wikipedia results after using the tool.",
    },
    {
        "category" : "Search",
        **YahooFinanceTool.get_tool_info(),
        "output": "YahooFinance results after using the tool.",
    },
    {
        "category" : "Search",
        **SerperSpecificSearchTool.get_tool_info(),
        "output": "SerperSpecific results after using the tool.",
    },
    {
        "category" : "Search",
        **SerperIntermediateSearchTool.get_tool_info(),
        "output": "SerperIntermediate results after using the tool.",
    },
    {
        "category" : "Search",
        **exaSearchTool.get_tool_info(),
        "output": "ExaSearch results after using the tool.",
    },
]
TOOLS_DICT_MAPPING = {tool["tool_name"]: tool for tool in TOOLS_DICT} # This is created for faster lookups when using tools in the agents
BASE_URL= getYamlAttribute('BASE_URL')
DEPLOYMENT_NAME= getYamlAttribute('DEPLOYMENT_NAME')
MODEL_NAME= getYamlAttribute('MODEL_NAME')
OPENAI_API_VERSION= getYamlAttribute('OPENAI_API_VERSION')
API_KEY = os.environ.get('AZURE_OPENAI_API_KEY')
llm_1 = AzureChatOpenAI(azure_deployment=DEPLOYMENT_NAME,
model_name = MODEL_NAME,
openai_api_version=OPENAI_API_VERSION,
openai_api_key=API_KEY,
azure_endpoint=BASE_URL)
serper_api_key = os.environ.get('SERPER_API_KEY')
os.environ["SERPER_API_KEY"] = serper_api_key

#verified
def getDuckduckgoSearchResults(searchString):
     logging.debug(searchString)
     search_tool = DuckDuckGoSearchRun()
     result=search_tool.run(searchString)
     logging.debug(f'duckduck output:{result}')
     return result


def getYoutubeSearchResults(searchString):
    tool = YouTubeSearchTool()
    result = tool.run(searchString)
    return result
    

def getSerperScrapeForSpecificTypeAndPeriod(searchString, type='news', tbs='qdr:y'):
    search = GoogleSerperAPIWrapper(type=type, tbs=tbs)
    logging.debug(f"searchString: {searchString}, type: {type}, tbs: {tbs}")
    output=search.results(searchString)
    pprint.pp(output)
    return output

def googleSerperSearchIntermediateQuestions(searchString):
    llm = llm_1 
    search = GoogleSerperAPIWrapper()
    tools = [
        Tool(
            name="Intermediate Answer",
            func=search.run,
            description="useful for when you need to ask with search",
        )
    ]
    self_ask_with_search = initialize_agent(
        tools, llm, agent=AgentType.SELF_ASK_WITH_SEARCH, verbose=False
    )
    results=self_ask_with_search.run(searchString)
    finalResult = pprint.pp(results)
    logging.debug(finalResult)
    return results

def googleFinanceStockSearch(searchString):
    os.environ["SERPAPI_API_KEY"]= os.environ.get('SERPER_API_KEY')
    os.environ["SERP_API_KEY"]= os.environ.get('SERPER_API_KEY')
    os.environ["AZURE_OPENAI_API_KEY"] = os.environ.get('AZURE_OPENAI_API_KEY')
    os.environ["AZURE_OPENAI_ENDPOINT"] = getYamlAttribute('BASE_URL')
    llm_2 = AzureChatOpenAI(
    model_name = MODEL_NAME,
    openai_api_version= getYamlAttribute('OPENAI_API_VERSION'),
    azure_deployment= getYamlAttribute('DEPLOYMENT_NAME'),
    )
    llm = llm_2 
    tools = load_tools(["google-scholar", "google-finance"], llm=llm)
    agent = initialize_agent(
    tools, llm, agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION, verbose=False
    )
    results= agent.run(searchString)
    finalResult= pprint.pp(results)
    logging.debug(finalResult)
    return results





def googleJobsSearch(input):
        os.environ["SERP_API_KEY"] = os.environ.get('SERP_API_KEY')
        tool = GoogleJobsQueryRun(api_wrapper=GoogleJobsAPIWrapper())
        results= tool.run(input)
        pprint.pp(results)
        logging.debug(results)
        return results

def yahooFinanceNews(searchString):
    llm = llm_1 
    tools = [YahooFinanceNewsTool()]
    tool = initialize_agent(
        tools,
        llm,
        handle_parsing_errors=True,
        agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION,
        verbose=False,
    )
    results= tool.run(searchString)
    pprint.pp(results)
    logging.debug(results)
    return results


def wikipediaTool(searchString):
    api_wrapper = WikipediaAPIWrapper(top_k_results=2, doc_content_chars_max=10000)
    tool = WikipediaQueryRun(api_wrapper=api_wrapper)
    results= tool.run(searchString)
    pprint.pp(results)
    logging.debug(results)
    return results
  
 
# def O365(searchString):
#     toolkit = O365Toolkit()
#     tools = toolkit.get_tools()
#     llm = llm_1
#     agent = initialize_agent(
#     tools=toolkit.get_tools(),
#     llm=llm,
#     verbose=False,
#     agent=AgentType.STRUCTURED_CHAT_ZERO_SHOT_REACT_DESCRIPTION,
#     )
#     result = agent.invoke(searchString)
#     return result

def gmail_toolkit(searchString):
    credentials = getYamlAttribute('GMAIL_CREDS')
    toolkit = GmailToolkit()
    tools = toolkit.get_tools()
    base_prompt = hub.pull("langchain-ai/openai-functions-template")
    instructions = "Use the Gmail API to search for emails in your inbox and return the results."
    prompt = base_prompt.partial(instructions=instructions)
    llm = llm_1
    agent = create_openai_functions_agent(llm, toolkit.get_tools(), prompt)
    agent_executor = AgentExecutor(
    agent=agent,
    tools=toolkit.get_tools(),
    verbose=True,
    )
    result = agent_executor.invoke({"input" : searchString})
    return result

def open_weather_app(searchString):
    llm = llm_1
    os.environ["OPENWEATHERMAP_API_KEY"] = os.environ.get('OPENWEATHERMAP_API_KEY')
    tools = load_tools(["openweathermap-api"], llm)
    agent_chain = initialize_agent(
    tools=tools,
    llm=llm,
    agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION,
    verbose=True,
    )
    result = agent_chain.invoke(searchString)
    logging.debug(result)
    return result

def nasatool(searchString):
    llm = llm_1
    nasa = NasaAPIWrapper()
    toolkit = NasaToolkit.from_nasa_api_wrapper(nasa)
    agent = initialize_agent(
        toolkit.get_tools(), llm, agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION, verbose=True
    )
    result = agent.invoke(searchString)   
    logging.debug(result)
    return result

def github_toolkit(searchString):
    os.environ["GITHUB_APP_ID"] = os.environ.get('GITHUB_APP_ID')
    os.environ["GITHUB_APP_PRIVATE_KEY"] = getYamlAttribute('GITHUB_APP_PRIVATE_KEY')
    os.environ["GITHUB_REPOSITORY"] = getYamlAttribute('GITHUB_REPOSITORY') 
    llm = llm_1
    github = GitHubAPIWrapper()
    toolkit = GitHubToolkit.from_github_api_wrapper(github)
    tools = toolkit.get_tools()
    agent = initialize_agent(
        tools=tools, llm=llm, agent=AgentType.STRUCTURED_CHAT_ZERO_SHOT_REACT_DESCRIPTION, verbose=True
    )
    result = agent.run(searchString)
    logging.debug(result)
    return result

# def powerbi(searchString):
#     llm = llm_1
#     toolkit = PowerBIToolkit(
#     powerbi=PowerBIDataset(
#         dataset_id="ca5fd8b4-6ce5-4d29-bca6-3871473c5013",
#         table_names=["Table"],
#         credential=DefaultAzureCredential(),
#     ),
#     llm=llm,
#     )
#     agent_executor = create_pbi_agent(
#     llm=llm,
#     toolkit=toolkit,
#     verbose=True,
#     )
#     result = agent_executor.run(searchString)
#     return result

def xorbits_toolkit(searchString):
    xorbotsCSVFileName = getYamlAttribute('xorbotsCSVFileName')
    data = pd.read_csv(xorbotsCSVFileName)
    llm = llm_1
    agent = create_xorbits_agent(llm, data, verbose=True, handle_parsing_errors=True)
    result = agent.run(searchString)
    logging.debug(result)
    return result

def sql_toolkit(searchString):
    sqlLiteDBName = getYamlAttribute('sqlLiteDBName')
    db = SQLDatabase.from_uri(sqlLiteDBName)
    llm = llm_1
    agent_executor = create_sql_agent(llm, db=db, agent_type="openai-tools", verbose=True)
    result = agent_executor.invoke(searchString)
    logging.debug(result)
    return result

def polygon_toolkit(searchString):
    os.environ["POLYGON_API_KEY"] = os.environ.get('POLYGON_API_KEY')
    llm = llm_1
    instructions = """You are an assistant."""
    base_prompt = hub.pull("langchain-ai/openai-functions-template")
    prompt = base_prompt.partial(instructions=instructions)
    polygon = PolygonAPIWrapper()
    toolkit = PolygonToolkit.from_polygon_api_wrapper(polygon)
    agent = create_openai_functions_agent(llm, toolkit.get_tools(), prompt)
    agent_executor = AgentExecutor(
    agent=agent,
    tools=toolkit.get_tools(),
    verbose=True,
    )
    result = agent_executor.invoke({"input" : searchString})
    logging.debug(result)
    return result

def DocuCompare(searchString):
    os.environ["AZURE_OPENAI_ENDPOINT"] = getYamlAttribute('BASE_URL')
    deployment_name = getYamlAttribute('EMBEDDING_DEPLOYMENT')
    llm = llm_1
    tools = []
    directory = getYamlAttribute('pdfFile')  # Specify the directory path here
    files = []
    for filename in os.listdir(directory):
        if filename.endswith(".pdf"):
            file_path = os.path.join(directory, filename)
            file_info = {"name": os.path.splitext(filename)[0], "path": file_path}
            files.append(file_info)
    for file in files:
        loader = PyPDFLoader(file["path"])
        pages = loader.load_and_split()
        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        docs = text_splitter.split_documents(pages)
        embeddings = AzureOpenAIEmbeddings(
            azure_deployment=deployment_name,
            openai_api_version=getYamlAttribute('OPENAI_API_VERSION'),
        )
        retriever = FAISS.from_documents(docs, embeddings).as_retriever()
        tools.append(
            Tool(
                name=file["name"],
                description=f"useful when you want to answer questions about {file['name']}",
                func=RetrievalQA.from_chain_type(llm=llm, retriever=retriever),
            )
        )
    agent = initialize_agent(
    agent=AgentType.OPENAI_FUNCTIONS,
    tools=tools,
    llm=llm,
    verbose=True,
    )
    result = agent({"input": searchString})
    logging.debug(result)
    return result

# def azure_cogs_toolkit(searchString):
#     os.environ["AZURE_COGS_KEY"] = os.environ.get('AZURE_COGS_KEY')
#     os.environ["AZURE_COGS_ENDPOINT"] = os.environ.get('AZURE_COGS_ENDPOINT')
#     os.environ["AZURE_COGS_REGION"] = os.environ.get('AZURE_COGS_REGION')
#     llm = llm_1
#     toolkit = AzureCognitiveServicesToolkit()
#     agent = initialize_agent(
#     tools=toolkit.get_tools(),
#     llm=llm,
#     agent=AgentType.STRUCTURED_CHAT_ZERO_SHOT_REACT_DESCRIPTION,
#     verbose=True,
#     )
#     result = agent.run(searchString)
#     return result

#utility function
def ExaAdvToolSetup22():
    
   # exa = Exa(api_key=os.environ["Exa_API_KEY"])
    exa = Exa(api_key=os.environ["EXA_API_KEY"])


    @tool
    def search(query: str, include_domains=None, start_published_date=None):
        """Search for a webpage based on the query.
        Set the optional include_domains (list[str]) parameter to restrict the search to a list of domains.
        Set the optional start_published_date (str) parameter to restrict the search to documents published after the date (YYYY-MM-DD).
        """
        return exa.search_and_contents(
            f"{query}",
            use_autoprompt=True,
            num_results=getYamlAttribute('EXA_NUM_SEARCH_RESULTS'), #2
            include_domains=include_domains,
            start_published_date=start_published_date,
        )

#utility function
def ExaAdvToolSetup():
    
   # exa = Exa(api_key=os.environ["Exa_API_KEY"])
    exa = Exa(api_key=os.environ["EXA_API_KEY"])


    @tool
    def search(query: str, include_domains=None, start_published_date=None):
        """Search for a webpage based on the query.
        Set the optional include_domains (list[str]) parameter to restrict the search to a list of domains.
        Set the optional start_published_date (str) parameter to restrict the search to documents published after the date (YYYY-MM-DD).
        """
        return exa.search_and_contents(
            f"{query}",
            use_autoprompt=True,
            num_results= getYamlAttribute('EXA_NUM_SEARCH_RESULTS'),
            include_domains=include_domains,
            start_published_date=start_published_date,
        )
    @tool
    def find_similar(url: str):
        """Search for webpages similar to a given URL.
        The url passed in should be a URL returned from `search`.
        """
        return exa.find_similar_and_contents(url, num_results=5)

    @tool
    def get_contents(ids: list[str]):
        """Get the contents of a webpage.
        The ids passed in should be a list of ids returned from `search`.
        """
        return exa.get_contents(ids)


    tools = [search, get_contents, find_similar]
    return tools
    
#utility function  
def ExaToolSetup():
    
    exa = Exa(api_key=os.environ["EXA_API_KEY"])


    @tool
    def search(query: str):
        """Search for a webpage based on the query."""
        return exa.search(f"{query}", use_autoprompt=True, num_results=5)


    @tool
    def find_similar(url: str):
        """Search for webpages similar to a given URL.
        The url passed in should be a URL returned from `search`.
        """
        return exa.find_similar(url, num_results=5)


    @tool
    def get_contents(ids: list[str]):
        """Get the contents of a webpage.
        The ids passed in should be a list of ids returned from `search`.
        """
        return exa.get_contents(ids)


    tools = [search, get_contents, find_similar]
    return tools
#working for small size results - can give issues when context length is higher than the maximm supported
def exaSearch(searchString):
    llm = llm_1
    tools = ExaAdvToolSetup()
    system_message = SystemMessage(
        content="You are a web researcher who answers user questions within the context length by looking up information on the internet and retrieving contents of helpful documents. Cite your sources."
    )

    agent_prompt = OpenAIFunctionsAgent.create_prompt(system_message)
    agent = OpenAIFunctionsAgent(llm=llm, tools=tools, prompt=agent_prompt)
    agent_executor = AgentExecutor(agent=agent, tools=tools, verbose=True)

    results= agent_executor.run(searchString)
    logging.debug(results)
    return results

def DallEImageGenerator(searchString):
    llm=llm_1
    tools = load_tools(["dalle-image-generator"])
    agent = initialize_agent(tools, llm, agent="zero-shot-react-description", verbose=False)
    image = agent.run(searchString)
    return image
        
#not working - to be implemented later
def text2Speech(searchString):
    llm = llm_1
    tools = load_tools(["eleven_labs_text2speech"])
    agent = initialize_agent(
      tools=tools,
      llm=llm,
      agent=AgentType.STRUCTURED_CHAT_ZERO_SHOT_REACT_DESCRIPTION,
      verbose=True,
    )
    audio_file = agent.run(searchString)
    tts = ElevenLabsText2SpeechTool()
    tts.stream_speech("this is bhaskar")
    #tts.play(audio_file)
    return audio_file

def test1():
      pat="What is the hometown of the reigning men's U.S. Open champion?"
      pat1="Collect and summarize recent news articles, press releases, and market analyses related to the stock  NOKIA and its industry"

      result= googleSerperSearchIntermediateQuestions(pat)
    
      result= getSerperScrapeForSpecificTypeAndPeriod("Tesla inc ", type="news", tbs="qdr:m")
      print(f"the result is: {result}")

      result = getSerperScrapeForSpecificTypeAndPeriod("Get best tourist places in India ", type="images", tbs="qdr:y")
            
      print(f"the result is: {result}")
      
      
def test2():
    # googleSerperSearch(pat1)
    # googleSerperSearch("Tesla inc ", type="news", tbs="qdr:m")
    # getSerperScrape("Tesla inc ", type="news", tbs="qdr:m")
    # O365("Could you search in my inbox and let me know if there's any mail?")        
    # gmail_toolkit("Create a new email with content on congraluations on your promotion to tanish@aiplanet.com and send it.")
    # open_weather_app("What is the weather in New York?")
    # nasatool("What is the latest news from NASA?")
    # print(github_toolkit("List all the code files and the content of files."))
    # powerbi("Describe Table")
    # xorbits_toolkit("What is the dataset about?")
    sql_toolkit("List the total sales per artist. Which artist's fans spent the most?")
    # polygon_toolkit("What is the last day's stock price of apple?")
    # playwright_toolkit("What are the headers on langchain.com?")   # requires async-await
    # DocuCompare("What is the difference between the two documents?")  
 


# Example Usage:
if __name__ == "__main__":


    role='Private Investment Advisor'
    goal="""Impress your customers with full analyses over stocks
        and completer investment recommendations"""
    backstory="""You're the most experienced investment advisor
        and you combine various analytical insights to formulate
        strategic investment advice. You are now working for
        a super important customer you need to impress."""
    task= """
            Collect and summarize recent news articles, press
            releases, and market analyses related to the stock and
            its industry.
            Pay special attention to any significant events, market
            sentiments, and analysts' opinions. Also include upcoming 
            events like earnings and others.
    
            Your final answer MUST be a report that includes a
            comprehensive summary of the latest news, any notable
            shifts in market sentiment, and potential impacts on 
            the stock.
            Also make sure to return the stock ticker.
            Make sure to use the most recent data as possible.
            Selected company by the customer:Nokia"""
            
    
    
 
          
    #wikipediaTool()
    Input= role + goal + backstory + task
    jobs= "give me an entry level job posting related to physics"
    #openAISearchJobs(jobs)
    #results= googleJobsSearch(jobs)
    # result= googleFinanceStockSearch("please get stock price of Nokia")
    #print(f"the result is: {results}")
    #tool = GoogleJobsQueryRun(api_wrapper=GoogleJobsAPIWrapper())

    #results=tool.run("Can I get an entry level job posting related to physics")
    #rl=pprint.pp(results)
    #jj="give me an entry level job posting related to physics"
   # googleJobsSearch(jj)
    #print(kk)
    #googleFinance("what is Nokia stock")
    
    pat2=" Can you please summarise only sachin tendulkar  world records"
    pat3= "How does Microsoft feels today comparing with Nvidia?"
    pat4="What happens to Nokia stocks today "
    pat5 = "Tell me a joke and read it out for me."
    results=wikipediaTool(pat2)
    print(f"the result wikipedia is: {results}")
    results = yahooFinanceNews(pat4)
    print(f"the result yahooFinanceNews is: {results}")

    results= getSerperScrapeForSpecificTypeAndPeriod(searchString="Nokia", type="news",tbs="qdr:m" )
    print(f"the result getSerperScrapeForSpecificTypeAndPeriod is: {results}")

    
    pat6="Summarize for me a fascinating article about cats."
    pat7= "Summarize for me an interesting article about AI from lesswrong.com published in  2024 not exeeding 2000 words."
    pat8="search for latest trends in Covid-19 and Cancer treatment that includes medicines, physical exercises, overall management and prevention aspects"
    # audio_file= text2Speech(pat5)
    pat9="Create an image of a halloween night at a haunted museum"
    results=getDuckduckgoSearchResults(pat8)
    #results= exaSearch(pat8)
    
    # image= DallEImageGenerator(pat9)
    print(f"the result is: {results}")
    # test1()
    # test2()
#Tool_description= ["getSerperScrape", "getSerperScrape:description"]



import os
import importlib
import json
import logging
import re
from typing import Dict, List
from langchain.chains import LLMChain
from langchain_core.prompts import PromptTemplate
from langchain_openai import AzureChatOpenAI
from openai import OpenAI
from langchain_core.messages import HumanMessage
from langchain_community.llms import Ollama
from genai_agents.utils.yamlParse import getYamlAttribute

BASE_URL = getYamlAttribute('BASE_URL') 
DEPLOYMENT_NAME = getYamlAttribute('DEPLOYMENT_NAME')   
MODEL_NAME= getYamlAttribute('MODEL_NAME')
OPENAI_API_VERSION =getYamlAttribute('OPENAI_API_VERSION')

LLM_TASK_API_KEY=os.environ.get('AZURE_OPENAI_API_KEY_PLANNER')
llm_1 = AzureChatOpenAI(
    azure_deployment=DEPLOYMENT_NAME,
    model_name=MODEL_NAME,  
    openai_api_version=OPENAI_API_VERSION, 
    openai_api_key=LLM_TASK_API_KEY,
    azure_endpoint=BASE_URL,
)

def extract_json_from_string(text_block):
    # Load JSON string into Python object
    json_extracted = text_block.replace("```json", "").replace("```", "")
    return json.loads(json_extracted)

def tools_handler(
    tools: List,
    task_input: Dict,
):
    model_prefix = """You are a smart assistant which is capable of smartly answering about how to solve a given problem in sequential tasks using the available tools.
                    The available tools are provided to you and you are supposed to answer What all tools are required to solve the problem and in which order with all of their parameters.
            """

    tool_prefix = """
        Tool describing format:
        [
            {
                "category" : "The category of the tool, defines what the tool can do in a basic manner.",
                "tool_name" : "The tool name",
                "description": "Description of the tool",
                "args": {"arg1": "arg1"},
                "cls": {
                    "kls": "Class name",
                    "module": "module name",
                },
                "output": "Output schema",
            }
        ]


        """

    tools_db = f"""
            The available tools:
            {tools}
            """
    task_desc = """
                    Task Statement:
            The Task statement has 5 parameters, and you must pay attention to all of them and give the best output.
            The 5 parameters are:
            "role" : This describes the role of the Language Model. This role tells you about your interest while looking for answers. For example if you are a private investment investor, then you will look for stock related information of different companies in recent timeframe using a search engine to properly answer the given task problem.
            "goal" : This is the goal which the final answer should achieve.
            "backstory" : This describes your role in a more detailed manner.
            "task" : The real task objective which needs to be solved. You are supposed to solve this problem by giving out sequential steps to solve the entire problem. 
            "instructions" : These are additional instructions which you have to follow while solving the task. It may be given or may not be given.
            "Tools_List" : These are the list of tools which you need to use. You are supposed to tell how to use these tools with all their paramters to solve the original problem.
            "OrderOfExecution" : This parameter tells you, whether you have to follow exact and strict order as given in the tools list or not while solving the main problem at hand. If the value is True we use the tools list in a strict manner, other wise a versatile manner.
            """

    task_input_str = ""

    for k, v in task_input.items():
        task_desc += f"\n{k}: {v}"

    example = (
        """
        Refer to some examples with the input information and the output information, you are supposed to answer the answer in this very format
        [
            {
                "role" : "Intelligent Question-answering Assistant",
                "goal" : "Answer the question with Best and very high accuracy with no mistake.",
                "backstory" : "You are an question-answering Assistant which has access to all the tools mentioned, like search and calculate, you are supposed to use thos to get the final answer.",
                "task" : "Where does the singer Arijit Singh belong to?",
                "instructions" : "",
                "Tools_list" : ["ScrapeAndSearchWebsite, SearchInternetGoogle, SearchInternetDuckDuck, getYoutubeSearchResults"],
                "OrderOfExecution" : true,
            }

            Tools:
            """
        + str(tools_db)
        + """
            Answer:
            ```json
                [
                    {
                        "category": "Search",
                        "tool_name": "DuckDuckGoSearch Tool",
                        "args": {"search_str": "trends in 2H 2023 onwards"},
                        "cls": {
                            "kls": "DuckDuckGoSearhTool",
                            "module": "genai_agents.tools.websearch.duckducksearch",
                        },
                        "output": "",
                    },
                    {
                        "category": "Search",
                        "tool_name": "Youtube Tool",
                        "args": {"search_str": "Trending in 2H 2023 onwards"},
                        "cls": {
                            "kls": "YoutubeSearchTool",
                            "module": "genai_agents.tools.websearch.youtubesearch",
                        },
                        "output": "",
                    },

                ]

            ```
        ]"""
    )

    suffix = "Provide a JSON-formatted response with the correct tool names and parameter values, without explanations, to solve the given problem accurately. Ensure the JSON format is error-free and follows the specified structure. Do not give any explanation paragraph, just give the required json answer. I want to use the final output in my code, so make sure it is in such a manner."

    final_prompt = f"""
    {model_prefix}
    {tool_prefix}
    {tools_db}
    {task_desc}
    "Task Input: {task_input_str}\n"
    {example}
    {suffix}
    """

    msg = HumanMessage(content=final_prompt)
    resp = llm_1([msg])
    json_resp_tool = extract_json_from_string(resp.content)

    tools_resp = []
    for tool in json_resp_tool:
        cls_name = tool["cls"]["kls"]
        mod_name = tool["cls"]["module"]
        cls = getattr(importlib.import_module(mod_name), cls_name)
        params = tool["args"]
        tool_obj = cls()._run(**params)
        tool_resp = {
            "tool_name": tool["tool_name"],
            "output": tool_obj,
        }
        tools_resp.append(tool_resp)

    # Concatenate tools_resp into a string
    # TODO: Make it return any kind of datatypes. Right now it supports strings only.
    tool_resp_str = ""
    for tool in tools_resp:
        tool_resp_str += f"{tool['tool_name']}: {tool['output']}\n"

    return tool_resp_str


def getEmail(inputString, role, backstory, goal, task, llm):
    template = "As a {role}, your primary objective is to {goal}. Your assigned task involves composing an email based on the provided summary. Here's some background information to guide you: {backstory}. Your challenge is to craft a detailed email within the given context. Please ensure that your email remains focused on the following context: {context}. Keep the email concise and relevant to the provided information."
    prompt = PromptTemplate(
        input_variables=["goal", "role", "task", "backstory", "context"],
        template=template,
    )
    prompt.format(
        context=inputString, role=role, backstory=backstory, goal=goal, task=task
    )
    chain = LLMChain(llm=llm, prompt=prompt)
    inputs = {
        "role": role,
        "backstory": backstory,
        "task": task,
        "goal": goal,
        "context": inputString,  # Assuming the inputString is the context
    }
    blog = chain.run(inputs)
    print(f"the blog is  {blog}")
    return blog

def getSummary(inputString, role, backstory, goal, task, llm):
    template = "As a {role}, your primary objective is to {goal}. Your assigned task involves performing the following {task}. Here's some background information to guide you: {backstory}. Your challenge is to craft a detailed summary within the given context. Please ensure that your summary remains focused on the following context:{context}. Keep the summary concise and relevant to the provided information.\n"
    prompt = PromptTemplate(
        input_variables=["goal", "role", "task", "backstory", "context"],
        template=template,
    )
    prompt.format(
        context=inputString, role=role, backstory=backstory, goal=goal, task=task
    )
    chain = LLMChain(llm=llm, prompt=prompt)
    inputs = {
        "role": role,
        "backstory": backstory,
        "task": task,
        "goal": goal,
        "context": inputString,  # Assuming the inputString is the context
    }
    blog = chain.run(inputs)
    print(f"output of getSummary {blog}")
    return blog

# function to generate code
def getCode(role, backstory, goal, task, llm):
    template = "As a {role}, your primary objective is to {goal}. Your assigned task involves performing the following {task} to the best of your ability. Here's some background information to guide you: {backstory}."
    prompt = PromptTemplate(
        input_variables=["goal", "role", "task", "backstory"], template=template
    )
    prompt.format(role=role, backstory=backstory, goal=goal, task=task)
    chain = LLMChain(llm=llm, prompt=prompt)
    inputs = {
        "role": role,
        "backstory": backstory,
        "task": task,
        "goal": goal,
        # "context": inputString  # Assuming the inputString is the context
    }
    code = chain.run(inputs)
    return code

# function to generate a review
def getReview(inputString, role, backstory, goal, task, llm):
    template = "As a {role}, your primary objective is to {goal}. Your assigned task involves performing the following {task}. Here's some background information to guide you: {backstory}. Your challenge is to perform the task within the given context:{context}. Keep the answer relevant to the provided information.\n"
    prompt = PromptTemplate(
        input_variables=["goal", "role", "task", "backstory", "context"],
        template=template,
    )
    prompt.format(
        context=inputString, role=role, backstory=backstory, goal=goal, task=task
    )
    chain = LLMChain(llm=llm, prompt=prompt)
    inputs = {
        "role": role,
        "backstory": backstory,
        "task": task,
        "goal": goal,
        "context": inputString,  # Assuming the inputString is the context
    }
    review = chain.run(inputs)
    return review

# Example: reuse your existing OpenAI setup

# def getFromLocalLLM(sysMsg, inputstring):

def handleLocalLLMTask(inputString, role, backstory, goal, task, llm):
    sysMsg = role + " " + backstory + " " + goal + " " + task
    # Point to the local server
    client = OpenAI(base_url="http://localhost:1236/v1", api_key="not-needed")

    completion = client.chat.completions.create(
        model="local-model",  # this field is currently unused
        messages=[
            {"role": "system", "content": sysMsg},
            {"role": "user", "content": inputString},
        ],
        temperature=0.7,
    )

    # print(completion.choices[0].message.content)
    return completion.choices[0].message.content

def getfromLocalLLM(inputString,role,backstory,goal,task,llm):
    llm_1=Ollama(model="llama2")
    template = "As a {role}, your primary objective is to {goal}. Your assigned task involves performing the following {task}. Here's some background information to guide you: {backstory}. Your challenge is to perform the task within the given context:{context}. Keep the answer relevant to the provided information.\n"
    prompt=PromptTemplate(input_variables=["goal","role","task","backstory","context"],
                          template=template)
    prompt.format(context=inputString,role=role,backstory=backstory,goal=goal,task=task)
    chain=LLMChain(llm=llm_1,prompt=prompt)
    inputs = {
        "role": role,
        "backstory": backstory,
        "task": task,
        "goal": goal,
        "context": inputString  # Assuming the inputString is the context
    }#inputstring with localllm
    review=chain.run(inputs)
    return review

def handleLLMTask(inputString, role, backstory, goal, task, llm):
    if llm == "local":
        return handleLocalLLMTask(inputString, role, backstory, goal, task, llm)
    elif llm == "gpt4" or llm == "openai ":
        print("not implemented for gpt4 and openai")
    elif llm == "azure" and role == "Coder":
        return getCode(role, backstory, goal, task, llm_1)
    elif llm == "azure" and role == "Reviewer":
        return getReview(inputString, role, backstory, goal, task, llm_1)
    elif llm == "azure" and role == "SUMMARISER":
        return getSummary(inputString, role, backstory, goal, task, llm_1)
    elif llm == "azure" and role == "EMAILER":
        return getEmail(inputString, role, backstory, goal, task, llm_1)
    elif llm == "azure":
        return getReview(inputString, role, backstory, goal, task, llm_1)
    elif (llm=="local-ollama"):
        return getfromLocalLLM(inputString,role,backstory,goal,task,llm)
    else:
        print("unsupported LLM")





# Example Usage:
if __name__ == "__main__":
    backstory="You are a software developer working on a platform that provides code snippets for various programming tasks."
    role="Coder"
    goal="To generate Python code for calculating the factorial of a number in a way that it can be easily copied and pasted by users for execution.Do not use any python libraries for it, hardcode it completely"
    task=" Develop a Python script that calculates the factorial of a given number, presenting the code in a line-by-line format to ensure clarity and ease of use for users."
    inputString="write program with given instructions"
    llm="abc"
   # ret= getfromLocalLLM(inputString,role,backstory,goal,task,llm)
    ret=handleLocalLLMTask(inputString, role, backstory, goal, task, llm)
    print(ret)


from  genai_agents.tools.tools_db import getDuckduckgoSearchResults
import logging
def search_string_in_list(my_list, target_string):
    """
    Search for a string in a list.

    Parameters:
    - my_list (list): The list to search through.
    - target_string (str): The string to search for.

    Returns:
    - list: A list of indices where the target string is found.
    """
    indices = []
    for index, item in enumerate(my_list):
        if item == target_string:
            indices.append(index)
    return indices


def getToolExecutionResults(agentName, role, goal, backstrory, task, capability, tools_list, llm_api):
    if (capability == "search_executor"):
        indices = search_string_in_list(tools_list, "duckduckgo-search")
        if indices:
             results = getDuckduckgoSearchResults(goal + task)
             print(f"executed duckduck search and results :{results}")
             return results
    else:
        logging.error("ERROR:tool_execution to be implemtned")
        print("ERROR in getToolExecutionResults\n")
        return "ERROR"
    



from genai_agents.tools.base import BaseTool, tool
from pydantic import BaseModel, Field


class NasaInputSchema(BaseModel):
    search_str: str = Field(description="Search string to be passed to the input.")


class NasaOutputSchema(BaseModel):
    response: str = Field(description="Response from the agent regarding action performed by Nasa.")


class NasaSearchTool(BaseTool):
    name: str = "NasaSearch Tool"
    description: str = (
        "A tool which can be used to retrieve information from Nasa's database like images, videos, documents, etc by using natural language."
    )

    @tool(args_schema=NasaInputSchema, output_schema=NasaOutputSchema)
    def _run(self, search_str: str = None):
        from genai_agents.tools.tools_db import nasatool

        return nasatool(searchString=search_str)
